#!/usr/bin/env python3
""" Fetch named test set(s)
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter

from . import fetcher


def get_parser():
    parser = ArgumentParser(description=__doc__,  # Usage from docstring
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('test_set_name', nargs='+',
                        help='Name(s) of test sets to fetch')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    for test_set in args.test_set_name:
        fetcher.get_set(test_set)


if __name__ == '__main__':
    main()



""" Load various data formats into xibabel images
"""

from pathlib import Path
import json
import logging
import importlib
import os.path as op

import numpy as np
import psutil

import nibabel as nib
from nibabel.spatialimages import HeaderDataError
from nibabel.filename_parser import parse_filename
from nibabel.fileholders import FileHolder

import fsspec
import xarray as xr
import dask.array as da


logger = logging.getLogger(__name__)


def max_available_div(div=10):
    """ Set default chunk as fraction of maximum aailable memory
    """
    return psutil.virtual_memory().available / div


MAXCHUNK_STRATEGY = max_available_div


class FDataObj:
    """ Wrapper for dataobj that returns floating point values from indexing.
    """

    def __init__(self, dataobj, dtype=np.float64):
        dtype = np.dtype(dtype)
        if not issubclass(dtype.type, np.inexact):
            raise ValueError(f'{dtype} should be floating point type')
        self._dataobj = dataobj
        self.dtype = dtype
        self.shape = dataobj.shape
        self.ndim = dataobj.ndim
        self.order = getattr(dataobj, 'order', None)

    def __getitem__(self, slicer):
        """ Return image data as floating point type ``self.dtype``.
        """
        return np.asanyarray(self._dataobj[slicer], dtype=self.dtype)

    def chunk_sizes(self, maxchunk=None):
        """ Calculate chunk sizes for dataobj shape

        Parameters
        ----------
        maxchunk : None or int, optional
            The largest allowable chunk sizes in bytes.

        Returns
        -------
        chunk_sizes : list
            Chunk sizes for Dask array creation, being number of elements in
            one chunk over all axes of array in ``self.dataobj``.
        """
        sizes = [None] * self.ndim
        if maxchunk is None:
            maxchunk = MAXCHUNK_STRATEGY()
        chunk_size = np.prod(self.shape) * self.dtype.itemsize
        if chunk_size <= maxchunk:
            return sizes
        axis_nos = range(self.ndim)
        if self.order == 'F':  # Assume C order by default.
            axis_nos = axis_nos[::-1]
        for axis_no in axis_nos:
            chunk_size //= self.shape[axis_no]
            n_chunks = maxchunk // chunk_size
            if n_chunks:
                sizes[axis_no] = int(n_chunks)
                return sizes
            sizes[axis_no] = 1
        return sizes


dim_recoder = {
    None: None,
    0: 'i',
    1: 'j',
    2: 'k'}


time_unit_scaler = {
    'sec': 1,
    'msec': 1 / 1000,
    'usec': 1 / 1_000_000}


class NiftiWrapper:

    def __init__(self, header):
        self.header = nib.Nifti1Header.from_header(header)

    def get_dim_labels(self):
        freq_dim, phase_dim, slice_dim = self.header.get_dim_info()
        return {'xib-FrequencyEncodingDirection': dim_recoder[freq_dim],
                'PhaseEncodingDirection': dim_recoder[phase_dim],
                'SliceEncodingDirection': dim_recoder[slice_dim]}

    def get_slice_timing(self):
        hdr = self.header
        freq_dim, phase_dim, slice_dim = hdr.get_dim_info()
        if not slice_dim:
            return None
        duration = hdr.get_slice_duration()
        if duration == 0:
            return None
        slice_start, slice_end = hdr['slice_start'], hdr['slice_end']
        n_slices = hdr.get_n_slices()
        if slice_start != 0 or slice_end != n_slices - 1:
            return None
        try:
            return list(hdr.get_slice_times())
        except HeaderDataError:
            return None

    def get_repetition_time(self):
        hdr = self.header
        zooms = hdr.get_zooms()
        if len(zooms) < 4:
            return None
        time_zoom = zooms[3]
        space_units, time_units = hdr.get_xyzt_units()
        if time_units == 'unknown':
            return None
        return time_unit_scaler[time_units] * time_zoom

    def get_affines(self):
        """ Collect valid affines
        """
        hdr = self.header
        affines = {}
        for affine_type in 'qform', 'sform':
            code = hdr.get_value_label(affine_type + '_code')
            if code != 'unknown':
                affine = getattr(hdr, 'get_' + affine_type)()
                affines[code] = affine.tolist()
        return affines

    def to_meta(self):
        meta = self.get_dim_labels()
        meta['SliceTiming'] = self.get_slice_timing()
        meta['RepetitionTime'] = self.get_repetition_time()
        meta['xib-affines'] = self.get_affines()
        return {k: v for k, v in meta.items() if v is not None}


def wrap_header(header):
    # We could try extracting more information from other file types, but
    return NiftiWrapper(header)


def load_zarr(url_or_path, **kwargs):
    r""" Load image from Zarr/Xibabel-format data at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `xr.open_dataarray`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    return xr.open_dataarray(url_or_path,
                             engine='zarr',
                             chunks='auto',
                             **kwargs)


class XibError(Exception):
    """ Errors from Xibabel file inference and operations
    """


class XibFormatError(XibError):
    """ Errors from Xibabel format specification or inference
    """


class XibFileError(XibError):
    """ Error from Xibabel file operations
    """


_JSON_MARKER = '__json__'


def _json_attrs2attrs(attrs):
    out = {}
    for key, value in attrs.items():
        if (isinstance(value, list) and
            len(value) == 2 and
            value[0] == _JSON_MARKER):
            value = json.loads(value[1])
        out[key] = value
    return out


def _1d_arrayable(v):
    try:
        arr = np.array(v)
    except ValueError:
        return False
    return arr.ndim < 2


def _attrs2json_attrs(attrs):
    out = {}
    for key, value in attrs.items():
        if (isinstance(value, dict) or
            (isinstance(value, (list, tuple)) and not _1d_arrayable(value))):
            value = [_JSON_MARKER, json.dumps(value)]
        out[key] = value
    return out


def _check_netcdf():
    if importlib.util.find_spec('h5netcdf') is None:
        raise XibFileError('Please install h5netcdf module to load netCDF')


def load_netcdf(url_or_path, **kwargs):
    r""" Load image from netCDF/Xibabel-format data at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `xr.open_dataarray`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    _check_netcdf()
    with fsspec.open(url_or_path) as fobj:
        img = xr.open_dataarray(fobj, engine='h5netcdf', **kwargs)
    img.attrs = _json_attrs2attrs(img.attrs)
    return img


def load(url_or_path, *, format=None, **kwargs):
    r""" Load image Xibabel or Nibbel image at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    format : None or str, optional
        If None, infer format from `url_or_path`, otherwise one of
        {', '.join(f'{e}' for e in PROCESSORS.format_processors)}
    \*\*kwargs : dict
        Any remaining named arguments passed to backend corresponding to
        `format`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    if format is None:
        format = PROCESSORS.guess_format(url_or_path)
    return PROCESSORS.get_loader(format)(url_or_path, **kwargs)


# Extensions we will search for valid images paired with JSON files.
_JSON_PAIRED_EXTS = ('.nii.gz', '.nii')


def drop_suffix(in_path, suffix):
    """ Drop suffix in `suffixes` from str or ``Path`` `in_path`

    `suffix` can be ``str`` (suffix to drop) or sequence.  If sequence, drops
    first matching suffix in sequence `suffix`.

    Parameters
    ----------
    in_path : str or Path
        Path from which to drop suffix.
    suffix : str or sequence
        If ``str``, suffix to drop.  If sequence, search for each suffix, and
        drop first found suffix from `in_path`.

    Returns
    -------
    out_path : str or Path
        Return type matches that of `in_path`.
    """
    if not hasattr(in_path, 'is_file'):
        return _drop_suffix_str(in_path, suffix)
    return in_path.with_name(_drop_suffix_str(in_path.name, suffix))


def _drop_suffix_str(path_str, suffix):
    suffixes = (suffix,) if isinstance(suffix, str) else suffix
    for suffix in suffixes:
        if path_str.endswith(suffix):
            return path_str[:-(len(suffix))]
    return path_str


def replace_suffix(in_path, old_suffix, new_suffix):
    """ Replace `in_path` suffix with `new_suffix`, allowing for `old_suffix`

    Always replace suffix of `in_path`, if present, but allowing for any
    special (multi-dot) suffixes in `old_sequence`.

    `old_suffix` can be ``str`` (suffix to drop) or sequence.  If sequence,
    drops first matching suffix in sequence `suffixes` before appending
    `new_suffix`.

    Parameters
    ----------
    in_path : str or Path
        Path from which to replace suffix.
    old_suffix : str or sequence
        If ``str``, suffix to drop before replacing.  If sequence, search for
        each suffix, and drop first found suffix from `in_path`.
    new_suffix : str
        Suffix to append.

    Returns
    -------
    out_path : str or Path
        Return type matches that of `in_path`.
    """
    is_path = hasattr(in_path, 'is_file')
    path_str = in_path.name if is_path else in_path
    dropped = _drop_suffix_str(path_str, old_suffix)
    replaced = op.splitext(dropped)[0] + new_suffix
    return in_path.with_name(replaced) if is_path else replaced


def _valid_or_raise(fs, url_base, exts):
    for ext in exts:
        target_url = replace_suffix(url_base, (), ext)
        if fs.exists(target_url):
            return fsspec.open(target_url, compression='infer')
    msg_suffix = (('one of ' if len(exts) > 1 else '') +
                  ', '.join(f"'{e}'" for e in exts))
    raise XibFileError(
        f"No valid file matching '{url_base}' + {msg_suffix}")


def load_bids(url_or_path, *, require_json=True, **kwargs):
    r""" Load image from BIDS-format data at `url_or_path`

    `url_or_path` may point directly to ``.json`` file, or to Nibabel format
    file, for which we expect a ``.json`` file to be present.

    In the second case, if `require_json` is True, and we cannot find a
    matching `.json` file, raise error, otherwise return image as best read.

    Parameters
    ----------
    url_or_path : str or Path
    require_json : {True, False}, optional, keyword-only
        If True, raise error if `url_or_path` is an image, and there is no
        matching JSON file.
    \*\*kwargs : dict
        Any remaining named arguments passed to `fsspec.open`.

    Returns
    -------
    bids_ximg : Xibabel image

    Raises
    ------
    XibFileError
        If `require_json` is True, `url_or_path` does not name a ``.json`` file
        and there is no ``.json`` file corresponding to `url_or_path`.
    """
    sidecar_file = None
    # If url_or_path has .json suffix, search for matching image file.
    if str(url_or_path).endswith('.json'):
        sidecar_file = fsspec.open(url_or_path, **kwargs)
        fs = sidecar_file.fs
        if not fs.exists(url_or_path):
            raise XibFileError(
                f'{url_or_path} does not appear to exist')
        img_file = _valid_or_raise(fs,
                                   drop_suffix(url_or_path, '.json'),
                                   _JSON_PAIRED_EXTS)
    else:  # Image file extensions.  Search for JSON.
        img_file = fsspec.open(url_or_path, compression='infer', **kwargs)
        fs = img_file.fs
        if not fs.exists(url_or_path):
            raise XibFileError(
                f'{url_or_path} does not appear to exist')
        url_uncomp = drop_suffix(url_or_path, _comp_exts())
        url_json = replace_suffix(url_uncomp, (), '.json')
        if fs.exists(url_json):
            sidecar_file = fs.open(url_json)
        elif require_json:
            raise XibFileError(
                f'BIDS loading {url_or_path} but no corresponding '
                f'{url_json} file, and `require_json` is True')
    img, meta = _nibabel2img_meta(img_file)
    if sidecar_file:
        with sidecar_file as fobj:
            meta.update(json.load(fobj))
    return _img_meta2ximg(img, meta, url_or_path)


def load_nibabel(url_or_path, **kwargs):
    r""" Load image from Nibabel-format data at `url_or_path`

    `url_or_path` may point directly to ``.json`` file, or to Nibabel format
    file, for expect a ``.json`` file to be present, in which case we will load
    it.

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `fsspec.open`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    return load_bids(url_or_path, require_json=False, **kwargs)


def _comp_exts():
    return tuple(f'.{ext}' for ext in fsspec.utils.compressions)


def _path2class(filename):
    compression_exts = _comp_exts()
    for klass in nib.all_image_classes:
        base, ext, gzext, ftype = parse_filename(filename,
                                                 klass.files_types,
                                                 compression_exts)
        if ftype == 'image':
            return klass
    raise XibFileError(f'No single-file Nibabel class for {filename}')


class FSFileHolder(FileHolder):

    def __del__(self):
        self.fileobj.close()


def _nibabel2img_meta(img_file):
    # Identify relevant files from img_file
    # Make file_map with opened files.
    if 'local' in img_file.fs.protocol:
        img = nib.load(img_file.path)
    else:  # Not local - use stream interface.
        img_klass = _path2class(img_file.full_name)
        # We are passing out opened fsspec files.
        fh = FileHolder(img_file.full_name, img_file.open())
        img = img_klass.from_file_map({'image': fh})
    return img, wrap_header(img.header).to_meta()


def _img_meta2ximg(img, meta, url_or_path):
    coords = {}
    if (TR := meta.get("RepetitionTime")):
        time_coords = np.arange(0, (img.shape[-1]) * TR, TR)
        coords = {
            "time":
            xr.DataArray(time_coords, dims=["time"], attrs={"units": "s"})}
    dataobj = FDataObj(img.dataobj)
    return xr.DataArray(da.from_array(dataobj, chunks=dataobj.chunk_sizes()),
                        dims=["i", "j", "k", "time"][:dataobj.ndim],
                        coords=coords,
                        name=_url2name(url_or_path),
                        # NB: zarr can't serialize numpy arrays as attrs
                        attrs=meta)


def _url2name(url_or_path):
    name = drop_suffix(url_or_path, _comp_exts())
    return Path(name).stem


def save(obj, url_or_path, format=None, **kwargs):
    r""" Save Xibabel image at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    format : None or str, optional
        If None, infer format from `url_or_path`, otherwise one of
        'zarr', 'netcdf', 'bids', 'nibabel'.
    \*\*kwargs : dict
        Any remaining named arguments passed to backend corresponding to
        `format`.

    Returns
    -------
    ret : object
        Return depends on `format`.  See `save_zarr` and `save_netcdf` for
        examples.
    """
    if format is None:
        format = PROCESSORS.guess_format(url_or_path)
    if format is None:
        raise XibFormatError(f"Could not guess format from {url_or_path}")
    return PROCESSORS.get_saver(format)(obj, url_or_path, **kwargs)


def save_zarr(obj, file_path, **kwargs):
    r""" Save Xibabel image at `url_or_path` in Zarr / Xibabel format.

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Arguments to pass to `to_zarr` method of `obj`.

    Returns
    -------
    zbe : ZarrBackendStore
    """
    return obj.to_zarr(file_path, mode='w', **kwargs)


def save_netcdf(obj, file_path, **kwargs):
    r""" Save Xibabel image at `url_or_path` in netCDF / Xibabel format.

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Arguments to pass to `to_netcdf` method of `obj`.

    Returns
    -------
    None
    """
    _check_netcdf()
    out = obj.copy()  # Shallow copy by default.
    out.attrs = _attrs2json_attrs(out.attrs)
    return out.to_netcdf(file_path, engine='h5netcdf', **kwargs)


class Processors:

    format_processors = {
        'zarr': dict(exts=('ximg',),
                     loader=load_zarr,
                     saver=save_zarr),
        'netcdf': dict(exts=('nc',),
                     loader=load_netcdf,
                     saver=save_netcdf),
        'bids': dict(exts=('json',),
                     loader=load_bids,
                     saver=None),
        'nibabel': dict(exts=(),  # Defer to Nibabel for extensions
                     loader=load_nibabel,
                     saver=None),
    }

    def __init__(self):
        self.loaders = set()
        self.savers = set()
        self.ext2fmt = {}
        for fmt, info in self.format_processors.items():
            if info['loader']:
                self.loaders.add(fmt)
            if info['saver']:
                self.savers.add(fmt)
            for ext in info['exts']:
                self.ext2fmt[ext] = fmt

    def get_loader(self, fmt):
        return self.get_processor(fmt, 'loader')

    def get_saver(self, fmt):
        return self.get_processor(fmt, 'saver')

    def get_processor(self, fmt, ptype='loader'):
        fmt = 'nibabel' if fmt is None else fmt
        pset = (self.loaders if ptype == 'loader'
                else self.savers)
        if fmt not in pset:
            raise XibFormatError(
                f"Cannot use format '{fmt}' as {ptype} for image; "
                f"valid formats are {','.join(pset)}")
        return self.format_processors[fmt][ptype]

    def guess_format(self, file_path):
        suff = str(file_path).split('.')[-1]
        return self.ext2fmt.get(suff)


PROCESSORS = Processors()



""" Code utilities for xibabel package
"""

def merge(d1, d2):
    """ Recursive merge of dictionaries `d1` and `d2`

    Parameters
    ----------
    d1 : dict
    d2 : dict

    Returns
    -------
    dout : dict
        Recursive merge of `d1` and `d2`.  For each dictionry contained in
        `d1`, update keys from equivalent (if present) in `d2`, but
        recursively, so that sub-dictionaries are also merged (updated).

    Notes
    -----
    Initially from `https://stackoverflow.com/a/50441142/1939576`_
    """
    # Recursive dictionary merge (update).
    if isinstance(d1, dict) and isinstance(d2, dict):
        return {
            **d1, **d2,
            **{k: d1[k] if d1[k] == d2[k] else merge(d1[k], d2[k])
            for k in {*d1} & {*d2}}
        }
    return d2



""" Test dictionary merge code.
"""

from xibabel.xutils import merge


def test_merge():
    a = {"name": "john", "phone":"123123123",
     "owns": {"cars": "Car 1", "motorbikes": "Motorbike 1"}}
    b = {"name": "john", "phone":"123", "owns": {"cars": "Car 2"}}
    assert merge(a, b) == {
        "name": "john", "phone": "123",
        "owns": {"cars": "Car 2", "motorbikes": "Motorbike 1"}}
    a = {"name": "john", "phone":"123123123",
         "owns": {"cars": {"Car 1": 1000, "Car 2": 2000},
                  "motorbikes": {"Motorbike 1": 100}}}
    b = {"name": "john", "phone":"123", "other": 1,
         "owns": {"cars": {"Car 1": 10000, "Car 4": 3000},
                  "motorbikes": {"Motorbike 2": 200}}}
    assert merge(a, b) == {
        "name": "john", "phone": "123", "other": 1,
        "owns": {"cars": {"Car 1": 10000, "Car 2": 2000, "Car 4": 3000},
                 "motorbikes": {"Motorbike 1": 100, "Motorbike 2": 200}}}
    b = {"name": "james", "phone": {'mobile': "+1 510 277 8554"}, "other": 1,
         "owns": ['Car 1', 'Car 2']}
    assert merge(a, b) == {
        "name": "james", "phone": {'mobile': "+1 510 277 8554"}, "other": 1,
         "owns": ['Car 1', 'Car 2']}



# Pytest configuration

from pathlib import Path
import os
import sys
import requests
import socket
import shutil
import time

from xibabel.testing import (JC_EG_FUNC, JC_EG_FUNC_JSON, JC_EG_ANAT,
                             JC_EG_ANAT_JSON, JH_EG_FUNC)


import pytest
from xprocess import ProcessStarter


HERE = Path(__file__).parent
TEST_APP_PORT = 8999

TEST_FILES = (
    JC_EG_FUNC,
    JC_EG_FUNC_JSON,
    JC_EG_ANAT,
    JC_EG_ANAT_JSON,
    JH_EG_FUNC,
)


class URLGetter:

    def __init__(self, server_path):
        self.server_path = server_path

    def make_url(self, url):
        return f'http://localhost:{TEST_APP_PORT}/{url}'

    def get(self, url, *args, **kwargs):
        return requests.get(self.make_url(url), *args, **kwargs)

    def read_text(self, url):
        return self.get(url).text

    def read_bytes(self, url):
        return self.get(url, stream=True).content

    def _write_to(self, url, contents, meth_name):
        path = url.replace('/', os.path.sep)
        getattr(self.server_path / path, meth_name)(contents)
        return self.make_url(url)

    def write_bytes_to(self, url, contents):
        return self._write_to(url, contents, 'write_bytes')

    def write_text_to(self, url, contents):
        return self._write_to(url, contents, 'write_text')


@pytest.fixture
def fserver(xprocess, tmp_path_factory, scope='session'):
    # We need our own server rather than pytest-httpserver
    # in order to support range requests.

    server_path = tmp_path_factory.mktemp("files")

    class Starter(ProcessStarter):

        n_tries = 10
        try_sleep = 1

        def startup_check(self):
            # From: https://github.com/piotrekw/tornado-pytest
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            for i in range(self.n_tries):
                result = sock.connect_ex(('127.0.0.1', TEST_APP_PORT))
                if result == 0:
                    return True
                time.sleep(self.try_sleep)
            return False

        terminate_on_interrupt = True

        args = [sys.executable,
                str(HERE / 'run_server.py'),
                str(server_path),
                '-p', str(TEST_APP_PORT)]

    os.environ['APP_PORT'] = str(TEST_APP_PORT)
    xprocess.ensure("fserver", Starter, restart=True)

    for tf in TEST_FILES:
        shutil.copy2(tf, server_path)

    yield URLGetter(server_path)

    # clean up whole process tree afterwards
    xprocess.getinfo("fserver").terminate()



import sys

import numpy as np  # noqa:F401
import numpy.testing as npt

import nibabel as nib  # noqa:F401
import xibabel as xib  # noqa:F401

from nibabel.testing import data_path


EG_4D = data_path / 'example4d.nii.gz'


def bench_img_load():
    sys.stdout.flush()
    print("\nNibabel / Xibabel load")
    print("--------------------")
    nib_time = npt.measure(f'nib.load("{EG_4D}")', 10)
    print(f'Nibabel load {nib_time:6.2f}')
    xib_time = npt.measure(f'np.array(xib.load("{EG_4D}"))', 10)
    print(f'Xibabel load {xib_time:6.2f}')
    sys.stdout.flush()



""" Testing utilties and data

See :mod:`fetcher` for the data fetcher, and instructions for adding files.
"""

from pathlib import Path

from ..loaders import drop_suffix
from .fetcher import Fetcher


fetcher = Fetcher(config_path=Path(__file__).parent)
get_set = fetcher.get_set
get_file = fetcher.get_file
have_file = fetcher.have_file
DATA_PATH = fetcher.data_path
skip_without_file = fetcher.skip_without_file


# Example files
JC_EG_FUNC = (DATA_PATH / 'ds000009' / 'sub-07' / 'func' /
              'sub-07_task-balloonanalogrisktask_bold.nii.gz')
JC_EG_FUNC_JSON = drop_suffix(JC_EG_FUNC, ('.nii.gz',)).with_suffix('.json')
JC_EG_ANAT= (DATA_PATH / 'ds000009' / 'sub-07' / 'anat' /
             'sub-07_T1w.nii.gz')
JC_EG_ANAT_JSON = drop_suffix(JC_EG_ANAT, ('.nii.gz',)).with_suffix('.json')
JH_EG_FUNC = (DATA_PATH / 'ds000105' / 'sub-1' / 'func' /
              'sub-1_task-objectviewing_run-01_bold.nii.gz')



""" Test loaders
"""

from pathlib import Path
import os
from importlib.util import find_spec
import gzip
from itertools import product
import json
import shutil

import numpy as np

import nibabel as nib

from xibabel import loaders
from xibabel.loaders import (FDataObj, load_bids, load_nibabel, load, save,
                             PROCESSORS, _json_attrs2attrs, drop_suffix,
                             replace_suffix, _attrs2json_attrs, wrap_header,
                             _path2class, XibFileError)
from xibabel.xutils import merge
from xibabel.testing import (JC_EG_FUNC, JC_EG_FUNC_JSON, JC_EG_ANAT,
                             JC_EG_ANAT_JSON, JH_EG_FUNC, skip_without_file,
                             fetcher)

import pytest

rng = np.random.default_rng()


class FakeProxy:

    def __init__(self, array, order=None):
        self._array = array
        self.order = order
        self.ndim = array.ndim
        self.shape = array.shape

    def __getitem__(self, slicer):
        return self._array[slicer]

    def __array__(self):
        return self[:]


def test_fdataobj_basic():
    arr = np.arange(24).reshape((2, 3, 4))
    proxy = FakeProxy(arr)
    fproxy = FDataObj(proxy)
    assert fproxy.ndim == arr.ndim
    assert fproxy.shape == arr.shape
    assert fproxy.order is None
    assert np.all(fproxy[1, 2, :] == arr[1, 2, :])
    assert arr[1, 2, :].dtype == np.arange(2).dtype
    assert fproxy.dtype == np.dtype(np.float64)
    assert fproxy[1, 2, :].dtype == np.dtype(np.float64)


def test_chunking(monkeypatch):
    arr_shape = 10, 20, 30
    arr = rng.normal(size=arr_shape)
    fproxy = FDataObj(FakeProxy(arr))
    f64s = np.dtype(float).itemsize
    monkeypatch.setattr(loaders, "MAXCHUNK_STRATEGY", lambda : arr.size * f64s)
    assert fproxy.chunk_sizes() == [None, None, None]
    monkeypatch.setattr(loaders, "MAXCHUNK_STRATEGY",
                        lambda : arr.size * f64s - 1)
    assert fproxy.chunk_sizes() == [9, None, None]
    c_fproxy = FDataObj(FakeProxy(arr, order='C'))
    assert c_fproxy.chunk_sizes() == [9, None, None]
    f_fproxy = FDataObj(FakeProxy(arr, order='F'))
    assert f_fproxy.chunk_sizes() == [None, None, 29]
    monkeypatch.setattr(loaders, "MAXCHUNK_STRATEGY",
                        lambda : arr.size * f64s / 2 - 1)
    assert fproxy.chunk_sizes() == [4, None, None]
    assert c_fproxy.chunk_sizes() == [4, None, None]
    assert f_fproxy.chunk_sizes() == [None, None, 14]
    monkeypatch.setattr(loaders, "MAXCHUNK_STRATEGY",
                        lambda : arr.size * f64s / 10 - 1)
    assert fproxy.chunk_sizes() == [1, 19, None]
    assert c_fproxy.chunk_sizes() == [1, 19, None]
    assert f_fproxy.chunk_sizes() == [None, None, 2]
    monkeypatch.setattr(loaders, "MAXCHUNK_STRATEGY",
                        lambda : arr.size * f64s / 30 - 1)
    assert fproxy.chunk_sizes() == [1, 6, None]
    assert c_fproxy.chunk_sizes() == [1, 6, None]
    assert f_fproxy.chunk_sizes() == [None, 19, 1]


def out_back(img, out_path):
    if out_path.is_file():
        os.unlink(out_path)
    nib.save(img, out_path)
    img = nib.load(out_path)
    return img, wrap_header(img.header).to_meta()


def test_nibabel_tr(tmp_path):
    # Default image.
    arr = np.zeros((2, 3, 4))
    img = nib.Nifti1Image(arr, np.eye(4), None)
    out_path = tmp_path / 'test.nii'
    back_img, meta = out_back(img, out_path)
    exp_meta = {'xib-affines': {'aligned': np.eye(4).tolist()}}
    assert meta == exp_meta
    arr = np.zeros((2, 3, 4, 6))
    img = nib.Nifti1Image(arr, np.eye(4), None)
    back_img, meta = out_back(img, out_path)
    assert meta == exp_meta
    img.header.set_xyzt_units('mm', 'sec')
    back_img, meta = out_back(img, out_path)
    exp_meta.update({'RepetitionTime': 1.0})
    assert meta == exp_meta
    img.header.set_xyzt_units('mm', 'msec')
    back_img, meta = out_back(img, out_path)
    assert meta['RepetitionTime'] == 1 / 1000
    img.header.set_xyzt_units('mm', 'usec')
    back_img, meta = out_back(img, out_path)
    assert meta['RepetitionTime'] == 1 / 1_000_000


def test_nibabel_slice_timing(tmp_path):
    # Default image.
    arr = np.zeros((2, 3, 4, 5))
    img = nib.Nifti1Image(arr, np.eye(4), None)
    out_path = tmp_path / 'test.nii'
    back_img, meta = out_back(img, out_path)
    exp_meta = {'xib-affines': {'aligned': np.eye(4).tolist()}}
    assert meta == exp_meta
    img.header.set_dim_info(None, None, 1)
    back_img, meta = out_back(img, out_path)
    assert meta == merge(exp_meta, {'SliceEncodingDirection': 'j'})
    img.header.set_dim_info(1, 0, 2)
    back_img, meta = out_back(img, out_path)
    exp_dim = merge(exp_meta,
                    {'PhaseEncodingDirection': 'i',
                     'xib-FrequencyEncodingDirection': 'j',
                     'SliceEncodingDirection': 'k'})
    assert meta == exp_dim
    img.header.set_slice_duration(1 / 4)
    back_img, meta = out_back(img, out_path)
    assert meta == exp_dim
    img.header['slice_start'] = 0
    back_img, meta = out_back(img, out_path)
    assert meta == exp_dim
    img.header['slice_end'] = 3
    back_img, meta = out_back(img, out_path)
    assert meta == exp_dim
    img.header['slice_code'] = 4  # NIFTI_SLICE_ALT_DEC
    back_img, meta = out_back(img, out_path)
    exp_timed = exp_dim.copy()
    exp_timed['SliceTiming'] = [0.75, 0.25, 0.5, 0]
    assert meta == exp_timed


def test_guess_format():
    root = Path('foo') / 'bar' / 'baz.suff'
    for v, exp in ((root, None),
                   (root.with_suffix('.nii'), None),
                   (root.with_suffix('.json'), 'bids'),
                   (root.with_suffix('.ximg'), 'zarr'),
                   (root.with_suffix('.nc'), 'netcdf'),
                   (root.with_suffix('.foo'), None)):

        assert PROCESSORS.guess_format(v) == exp
        assert PROCESSORS.guess_format(str(v)) == exp


def test__path2class():
    for url, exp_class in (
        ('/foo/bar/sub-07_T1w.nii.gz', nib.Nifti1Image),
        ('http://localhost:8999/sub-07_T1w.nii.gz', nib.Nifti1Image),
        ('/foo/bar/sub-07_T1w.nii', nib.Nifti1Image),
        ('http://localhost:8999/sub-07_T1w.nii', nib.Nifti1Image),
        ('/foo/bar/sub-07_T1w.mnc', nib.Minc1Image),
        ('http://localhost:8999/sub-07_T1w.mnc', nib.Minc1Image),
    ):
        assert _path2class(url) == exp_class


def test_drop_suffix():
    for inp, suffixes, exp_out in (
        ('foo/bar', ['.nii'], 'foo/bar'),
        ('foo/bar', '.nii', 'foo/bar'),
        ('foo/bar.baz', ['.nii'], 'foo/bar.baz'),
        ('foo/bar.nii', ['.nii'], 'foo/bar'),
        ('foo/bar.nii', '.nii', 'foo/bar'),
        ('foo/bar.nii.gz', ['.nii'], 'foo/bar.nii.gz'),
        ('foo/bar.nii.gz', ['.nii.gz', '.nii'], 'foo/bar'),
    ):
        assert drop_suffix(inp, suffixes) == exp_out
        assert drop_suffix(Path(inp), suffixes) == Path(exp_out)


def test_replace_suffix():
    for inp, suffixes, new_suffix, exp_out in (
        ('foo/bar', ['.nii'], '.json', 'foo/bar.json'),
        ('foo/bar', '.nii', '.json', 'foo/bar.json'),
        ('foo/bar.baz', ['.nii'], '.boo', 'foo/bar.boo'),
        ('foo/bar.nii', ['.nii'], '.boo', 'foo/bar.boo'),
        ('foo/bar.nii', '.nii', '.boo', 'foo/bar.boo'),
        ('foo/bar.nii.gz', ['.nii'], '.boo', 'foo/bar.nii.boo'),
        ('foo/bar.nii.gz', ['.nii.gz', '.nii'], '.boo', 'foo/bar.boo'),
    ):
        assert replace_suffix(inp, suffixes, new_suffix) == exp_out
        assert replace_suffix(Path(inp), suffixes, new_suffix) == Path(exp_out)


def test_json_attrs():
    # Test utilities to load / save JSON attrs
    d = {'foo': 1, 'bar': [2, 3]}
    assert _attrs2json_attrs(d) == d
    assert _json_attrs2attrs(d) == d
    dd = {'foo': 1, 'bar': {'baz': 4}}
    ddj = {'foo': 1, 'bar': ['__json__', '{"baz": 4}']}
    assert _attrs2json_attrs(dd) == ddj
    assert _json_attrs2attrs(ddj) == dd
    arr = rng.integers(0, 10, size=(3, 4)).tolist()
    arr_j = json.dumps(arr)
    dd = {'foo': 1, 'bar': {'baz': [2, 3]}, 'baf': arr}
    ddj = {'foo': 1,
           'bar': ['__json__', '{"baz": [2, 3]}'],
           'baf': ['__json__', arr_j]}
    assert _attrs2json_attrs(dd) == ddj
    assert _json_attrs2attrs(ddj) == dd


@skip_without_file(JC_EG_FUNC)
def test_nib_loader_jc():
    img = nib.load(JC_EG_FUNC)
    ximg = load_nibabel(JC_EG_FUNC)
    assert ximg.attrs == JC_EG_FUNC_META
    assert np.all(np.array(ximg) == img.get_fdata())


@skip_without_file(JH_EG_FUNC)
def test_nib_loader_jh():
    img = nib.load(JH_EG_FUNC)
    ximg = load_nibabel(JH_EG_FUNC)
    assert ximg.attrs == {'RepetitionTime': 2.5,
                          'xib-affines':
                          {'scanner': img.affine.tolist()}
                         }


if fetcher.have_file(JC_EG_FUNC):
    img = nib.load(JC_EG_FUNC)
    JC_EG_FUNC_META = json.loads(JC_EG_FUNC_JSON.read_text())
    JC_EG_FUNC_META_RAW = {
        'xib-FrequencyEncodingDirection': 'i',
         'PhaseEncodingDirection': 'j',
         'SliceEncodingDirection': 'k',
         'RepetitionTime': 2.0,
         'xib-affines':
         {'scanner': img.affine.tolist()}
    }
    JC_EG_FUNC_META.update(JC_EG_FUNC_META_RAW)


if fetcher.have_file(JC_EG_ANAT):
    img = nib.load(JC_EG_ANAT)
    JC_EG_ANAT_META = json.loads(JC_EG_ANAT_JSON.read_text())
    JC_EG_ANAT_META_RAW = {
        'xib-FrequencyEncodingDirection': 'j',
         'PhaseEncodingDirection': 'i',
         'SliceEncodingDirection': 'k',
         'xib-affines':
         {'scanner': img.affine.tolist()}
    }
    JC_EG_ANAT_META.update(JC_EG_ANAT_META_RAW)


@skip_without_file(JC_EG_ANAT)
def test_anat_loader():
    img = nib.load(JC_EG_ANAT)
    for loader, in_path in product(
        (load, load_bids, load_nibabel),
        (JC_EG_ANAT, str(JC_EG_ANAT),
         JC_EG_ANAT_JSON, str(JC_EG_ANAT_JSON))):
        ximg = loader(in_path)
        assert ximg.shape == (176, 256, 256)
        assert ximg.name == JC_EG_ANAT.name.split('.')[0]
        assert ximg.attrs == JC_EG_ANAT_META
        assert np.all(np.array(ximg) == img.get_fdata())


@skip_without_file(JC_EG_ANAT)
def test_anat_loader_http(fserver):
    nb_img = nib.load(JC_EG_ANAT)
    # Read nibabel from HTTP
    # Original gz
    name_gz = JC_EG_ANAT.name
    # Uncompressed, no gz
    name_no_gz = JC_EG_ANAT.with_suffix('').name
    out_path = fserver.server_path / name_no_gz
    with gzip.open(JC_EG_ANAT, 'rb') as f:
        out_path.write_bytes(f.read())
    for name in (name_gz, name_no_gz):
        out_url = fserver.make_url(name)
        ximg = load(out_url)
        # Check we can read the data
        ximg.compute()
        # Check parameters
        assert ximg.shape == (176, 256, 256)
        assert ximg.name == JC_EG_ANAT.name.split('.')[0]
        assert ximg.attrs == JC_EG_ANAT_META
        assert np.all(np.array(ximg) == nb_img.get_fdata())


@skip_without_file(JC_EG_ANAT)
def test_anat_loader_http_params(fserver, tmp_path):
    # Test params get passed through in kwargs.
    nb_img = nib.load(JC_EG_ANAT)
    out_url = 'simplecache::' + fserver.make_url(JC_EG_ANAT.name)
    out_cache = tmp_path / 'files'
    assert not out_cache.is_dir()
    ximg = load(out_url,
                simplecache={'cache_storage': str(out_cache)})
    assert np.all(np.array(ximg) == nb_img.get_fdata())
    assert out_cache.is_dir()
    assert len(list(out_cache.glob('*'))) == 2  # JSON and Nifti


@skip_without_file(JC_EG_ANAT)
def test_round_trip(tmp_path):
    ximg = load(JC_EG_ANAT)
    assert ximg.shape == (176, 256, 256)
    out_path = tmp_path / 'out.ximg'
    save(ximg, out_path)
    back = load(out_path)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_META
    # And again
    save(ximg, out_path)
    back = load(out_path)
    assert back.attrs == JC_EG_ANAT_META
    # With url
    back = load(f'file:///{out_path}')
    assert back.attrs == JC_EG_ANAT_META


@pytest.mark.skipif(not find_spec('h5netcdf'),
                    reason='Need h5netcdf module for test')
@skip_without_file(JC_EG_ANAT)
def test_round_trip_netcdf(tmp_path):
    ximg = load(JC_EG_ANAT)
    out_path = tmp_path / 'out.nc'
    save(ximg, out_path)
    back = load(out_path)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_META
    back = load(f'file:///{out_path}')
    assert back.attrs == JC_EG_ANAT_META


def test_tornado(fserver):
    # Test static file server for URL reads
    fserver.write_text_to('text_file', 'some text')
    fserver.write_bytes_to('binary_file', b'binary')
    response = fserver.get('text_file')
    assert response.status_code == 200
    assert response.text == 'some text'
    assert fserver.read_text('text_file') == 'some text'
    assert fserver.read_bytes('text_file') == b'some text'
    assert fserver.read_bytes('binary_file') == b'binary'


@pytest.mark.skipif(not find_spec('h5netcdf'),
                    reason='Need h5netcdf module for test')
@skip_without_file(JC_EG_ANAT)
def test_round_trip_netcdf_url(fserver):
    ximg = load(JC_EG_ANAT)
    save(ximg, fserver.server_path / 'out.nc')
    out_url = fserver.make_url('out.nc')
    back = load(out_url)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_META


@skip_without_file(JC_EG_ANAT)
def test_matching_img_error(tmp_path):
    out_json = tmp_path / JC_EG_ANAT_JSON.name
    with pytest.raises(XibFileError, match='does not appear to exist'):
        load(out_json)
    shutil.copy2(JC_EG_ANAT_JSON, tmp_path)
    with pytest.raises(XibFileError, match='No valid file matching'):
        load(out_json)
    out_img = tmp_path / JC_EG_ANAT.name
    shutil.copy2(JC_EG_ANAT, tmp_path)
    back = load(out_img)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_META
    os.unlink(out_img)
    with pytest.raises(XibFileError, match='does not appear to exist'):
        load(out_img)
    shutil.copy2(JC_EG_ANAT, tmp_path)
    os.unlink(out_json)
    back = load(out_img)
    assert back.attrs == JC_EG_ANAT_META_RAW
    back = load_bids(out_img, require_json=False)
    assert back.attrs == JC_EG_ANAT_META_RAW
    with pytest.raises(XibFileError, match='`require_json` is True'):
        load_bids(out_img, require_json=True)



#!/usr/bin/env python3
""" Run Tornado static file server at given path
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
import asyncio

from tornado import web


def make_app(path):
    return web.Application([
        (r"/(.*)", web.StaticFileHandler, {"path": path})
    ])


async def run_server(path, port=8899):
    (path / 'server_marker').write_text('static server')
    app = make_app(path)
    app.listen(int(port))
    shutdown_event = asyncio.Event()
    await shutdown_event.wait()


def get_parser():
    parser = ArgumentParser(description=__doc__,  # Usage from docstring
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('files_path',
                        help='Path from which to serve files')
    parser.add_argument('-p', '--port', default='8899',
                        help='Port for server')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='If set, show messages')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.verbose:
        print(f'Serving {args.files_path} on port {args.port}')
    asyncio.run(run_server(Path(args.files_path), args.port))


if __name__ == '__main__':
    main()



""" Test scripting of  GLM
"""

import numpy as np

import pandas as pd

import xarray as xr

import nibabel as nib
import xibabel as xib

from xibabel import testing

# For constructing the design.
from nipy.modalities.fmri.design import block_design, natural_spline

import pytest

img_path_root = (testing.DATA_PATH /
                'ds000105' /
                'sub-1' /
                'func' /
                'sub-1_task-objectviewing_run-01_')

bold_path = img_path_root.with_name(img_path_root.name + 'bold.nii.gz')
tsv_path = img_path_root.with_name(img_path_root.name + 'events.tsv')


@pytest.mark.skipif(not bold_path.is_file(), reason=f'Missing "{bold_path}"')
@pytest.mark.skipif(not tsv_path.is_file(), reason=f'Missing "{tsv_path}"')
def test_glm():

    # Load the events ready to make a design.
    event_df = pd.read_csv(tsv_path, sep='\t')
    df = event_df.rename(columns={'onset': 'start', 'duration': 'end'})
    df['end'] = df['start'] + df['end']
    block_spec = df.to_records(index=None)

    nib_img = nib.load(bold_path)
    vol_shape = nib_img.shape[:-1]
    n_vols = nib_img.shape[-1]
    TR = nib_img.header.get_zooms()[-1]
    # Of course this array comes naturally from xirr['time'] below.
    t = np.arange(n_vols) * TR
    regressors, contrasts = block_design(block_spec, t)
    con_tt0 = contrasts['trial_type_0']
    n_contrasts = con_tt0.shape[0]
    # Add drift regressors.
    drift = natural_spline(t)
    n_drift = drift.shape[1]
    design_matrix = np.hstack([regressors, drift])
    # Contrasts for the eight event types.
    con_mat = np.hstack([con_tt0, np.zeros((n_contrasts, n_drift))])

    # For notation!
    X = design_matrix

    # Analysis the old way.
    # The 4D array (i, j, k, time)
    data = nib_img.get_fdata()

    # Reshape to time by (i * j * k)
    vols_by_voxels = np.reshape(data, (-1, n_vols)).T

    # Run estimation with B = X+ Y
    # To get B such that errors (E = Y - X @ B) are minimized in the sense
    # of having the smallest np.sum(E ** 2, axis=0) for each column.
    # Betas for each voxel (n_cols_in_X by n_voxels).
    pX = np.linalg.pinv(X)
    B = pX @ vols_by_voxels

    # Contrast applied.  Two slopes compared
    c = con_mat[4, :]
    con_vec = c @ B

    con_arr = np.reshape(con_vec, vol_shape)

    # OK - that was the crappy way.
    # Now the awesome way.
    xib_img = xib.load(bold_path)
    # assert isinstance(xib_img, xr.DataArray)

    # Make the design
    xesign = xr.DataArray(pX, dims=['p', 'time'])

    # Optional: make the data chunky.
    chunked = xib_img.chunk({'k': 5})

    # Do the estimation
    xB = xr.dot(xesign, chunked, dim=['time', 'time'])

    xC = xr.DataArray(c, dims=['p'])

    x_c_arr = xr.dot(xC, xB, dim=['p', 'p'])

    assert np.allclose(x_c_arr, con_arr)



""" Fetching and caching data from Datalad stores

We currently only have one data source — `https://www.datalad.org`_.

The datasets we are using come from the `OpenNeuro collection
<https://docs.openneuro.org/git>`_.

To add a file to the data collection, see `test_files.yml`.  Add a new entry
under an existing or new Datalad repository, following the examples.  Then
consider adding that file, with the repository name prefix, to the
`test_sets.yml` file.
"""

import os
from pathlib import Path
from subprocess import check_call
from collections import abc

import yaml


class TestFileError(Exception):
    """ Error for missing or damaged test file """


class Fetcher:

    def __init__(self,
                 config_path=Path(__file__).parent,
                 data_path=None,
                 files_config='test_files.yml',
                 sets_config='test_sets.yml',
                ):
        data_path = data_path if data_path else self.get_data_path()
        self.data_path = Path(data_path).expanduser().resolve()
        self.data_path.mkdir(parents=True, exist_ok=True)
        config_path = Path(config_path)
        self._files_config = self._get_config(files_config, config_path)
        self._sets_config = self._get_config(sets_config, config_path)
        self._file_sources = self._parse_configs()

    def _get_config(self, config, config_path):
        if isinstance(config, abc.Mapping):
            return config
        if isinstance(config, str):
            config = Path(config)
        if not config.is_absolute():
            config = config_path / config
        with open(config) as fobj:
            return yaml.load(fobj, Loader=yaml.SafeLoader)

    def _parse_configs(self):
        out = {}
        for source, info in self._files_config.items():
            if source == 'datalad':
                for repo in info:
                    url = repo['repo']
                    root = url.split('/')[-1]
                    for filename in repo['files']:
                        out[f'{root}/{filename}'] = {'type': 'datalad',
                                                     'repo': url}
            else:
                raise TestFileError(f'Do not recognize source "{source}"')
        return out

    def get_data_path(self):
        return os.environ.get('XIB_DATA_PATH', '~/.xibabel/data')

    def have_file(self, path):
        """ True if `path` is local file

        We need this function to deal with Datalad, which used ``git annex`` to
        store links to larger files.  This function needs to check whether the file
        we have is the actual file, or a link (sort-of-thing) to the file, that
        still needs a ``datalad get`` to fetch the actual contents.
        """
        if not path.is_file():  # Can be True for git annex links on Windows.
            return False
        if path.is_symlink():  # Appears to be True on Unices.
            return True
        # By exploration on Windows - detecting git annex placeholder for file.
        with open(path, 'rb') as fobj:
            start = fobj.read(15)
        return start != b'/annex/objects/'

    def _get_datalad_file(self, path_str, repo_url):
        path_str = self._source2path_str(path_str)
        file_path = (self.data_path / path_str).resolve()
        repo_str, file_str = path_str.split('/', 1)
        repo_path = self.data_path / repo_str
        if not repo_path.is_dir():
            check_call(['datalad', 'install', repo_url],
                       cwd=self.data_path)
        if not self.have_file(file_path):
            check_call(['datalad', 'get', file_str], cwd=repo_path)
        return file_path

    def _source2path_str(self, path_or_str):
        if isinstance(path_or_str, str):
            path_or_str = Path(path_or_str)
        if path_or_str.is_absolute():
            path_or_str = path_or_str.relative_to(self.data_path)
        return '/'.join(path_or_str.parts)

    def get_file(self, path_str):
        path_str = self._source2path_str(path_str)
        source = self._file_sources.get(path_str)
        if source is None:
            raise TestFileError(
                f'No record of "{path_str}" as test file')
        if source['type'] == 'datalad':
            return self._get_datalad_file(path_str, source['repo'])
        raise TestFileError(f'Unexpected source "{source}"')

    def get_set(self, set_name):
        out_paths = set()
        set_dict = self._sets_config.get(set_name)
        if set_dict is None:
            raise TestFileError(f'No set called "{set_name}"')
        for subset in set_dict.get('superset_of', []):
            out_paths = out_paths.union(self.get_set(subset))
        for path_str in set_dict.get('files', []):
            out_paths.add(self.get_file(path_str))
        return out_paths

    def skip_without_file(self, path):
        """ Pytest decorator to skip test if file not available
        """
        import pytest
        return pytest.mark.skipif(
            not self.have_file(path),
            reason=f'This test requires file {path}')
